Package: groupedHyperframe
Type: Package
Title: Grouped Hyper Data Frame
Version: 0.4.3
Date: 2026-06-20
Description: To aggregate a hyper data frame, defined in the package 'spatstat.geom', according to a grouping structure.  To facilitate downstream analysis based on a "grouped" hyper data frame.  The author has retired from academic research. Accordingly, this package should not be considered a validated tool for use in peer-reviewed publications or as the basis for grant applications.  Backward compatibility with user-code published in <doi:10.1093/bioinformatics/btaf430> is not maintained in versions >= 0.4.0 of this package. The authors of those publications are the appropriate contacts for reproducibility inquiries. 
LazyData: true
LazyDataCompression: xz
Encoding: UTF-8
License: GPL-2
Language: en-US
Authors@R: c(person(given = 'Tingting', family =
     'Zhan', role = c('aut', 'cre'), email =
     'tingtingzhan@gmail.com', comment = c(ORCID
     = '0000-0001-9971-4844')))
URL: https://github.com/tingtingzhan/groupedHyperframe
Depends: R (>= 4.6)
Imports: cli, spatstat.geom
Suggests: survival, spatstat.data
Config/roxygen2/version: 8.0.0
Config/roxygen2/markdown: TRUE
NeedsCompilation: no
Packaged: 2026-06-18 22:19:46 UTC; tingtingzhan
Author: Tingting Zhan [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9971-4844>)
Maintainer: Tingting Zhan <tingtingzhan@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-21 15:10:02 UTC
Built: R 4.6.0; ; 2026-06-21 15:39:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(aggregate,hyperframe)
S3method(aug4gam,data.frame)
S3method(aug4gam,hyperframe)
S3method(aug4gam,vectorlist)
S3method(print,vectorlist)
S3method(t,vectorlist)
export(aggregate2)
export(as.vectorlist)
export(aug4gam)
export(is.vectorlist)
export(pppBy)
import(cli)
importFrom(cli,cli_text)
importFrom(cli,col_green)
importFrom(cli,col_red)
importFrom(cli,style_bold)
importFrom(spatstat.geom,as.hyperframe.data.frame)
importFrom(spatstat.geom,cbind.hyperframe)
importFrom(spatstat.geom,names.hyperframe)
importFrom(spatstat.geom,owin)
importFrom(spatstat.geom,ppp)
importFrom(spatstat.geom,split.ppp)
importFrom(spatstat.geom,subset.hyperframe)
importFrom(stats,aggregate)
importFrom(stats,aggregate.data.frame)
importFrom(stats,model.frame.default)
importFrom(stats,setNames)
importFrom(stats,update.formula)

#' @title Handy `FUN`ctions for \link[stats]{aggregate}
#' 
#' @description
#' 
#' Handy functions to be passed to the parameter `FUN` of the function \link[stats]{aggregate}
#' and/or [aggregate2()].
#' 
#' @param x an R object
#' 
#' @returns
#' The function [unique_or_identity()] returns an R object.
#' 
#' @name aggregate_FUN
#' @export
unique_or_identity <- \(x) {
  u <- unique(x)
  if (length(u) == 1L) return(u)
  return(x)
}




unsimplify <- \(x) {
  # Surv is.matrix
  # Surv does not inherits from matrix :)
  if (!inherits(x, what = 'matrix')) return(x)
  x |> 
    asplit(MARGIN = 1L, drop = TRUE)
  # i.e., `simplify = TRUE` for Surv column,
  # but un-simplify for 'matrix'
}

#' @title Augment Hypercolumn(s) for \link[mgcv]{gam}
#' 
#' @description
#' Augment *all* \link[base]{numeric} \link[base]{vector} hypercolumns in a \link[spatstat.geom]{hyperframe} 
#' for \link[mgcv]{gam}.
#' 
#' @param x see **Usage**
#' 
#' @param ... additional parameters, currently of no use
#' 
#' @returns
#' The `S3` generic function [aug4gam()] returns a \link[base]{data.frame}.
#' 
#' @author 
#' Tingting Zhan, Erjia Cui
#' 
#' @name aug4gam
#' @export
aug4gam <- function(x, ...) UseMethod(generic = 'aug4gam')

#' @rdname aug4gam
#' @export
aug4gam.data.frame <- function(x, ...) return(x)


#' @rdname aug4gam
#' @export
aug4gam.hyperframe <- function(x, ...) {
  
  hc <- unclass(x)$hypercolumns
  
  id <- hc |> 
    vapply(FUN = is.vectorlist, mode = 'numeric', FUN.VALUE = NA)
  if (!any(id)) stop('no vector-list hypercolumn')

  z <- hc[id] |> 
    lapply(FUN = aug4gam.vectorlist) |>
    unlist(recursive = FALSE)
  
  dat <- unclass(x)$df
  # seems that only
  # base::`$<-.data.frame`
  # base::`[[<-.data.frame`
  # respects 'matrix'-columns
  
  for (i in seq_along(z)) {
    dat[[names(z)[i]]] <- z[[i]]
  }
  
  return(dat)
  
}

#' @rdname aug4gam
#' @export
aug4gam.vectorlist <- function(x, ...) {
  
  Y <- x |> 
    do.call(what = rbind)
  
  cnm <- Y |> 
    colnames()
  x. <- if (all(grepl(pattern = '%$', x = cnm))) {
    # returned from ?stats:::quantile.default
    tmp <- cnm |>
      gsub(pattern = '%$', replacement = '') |>
      as.double()
    tmp / 1e2
  } else cnm |> as.double() 
  
  if (!length(x.) || !is.numeric(x.) || anyNA(x.) || is.unsorted(x., strictly = TRUE)) {
    stop('must have names convertible to strictly-increasing numerics')
  }
  
  colnames(Y) <- x.
  
  X <- tcrossprod(rep(1, times = length(x)), x.)
  nc <- length(x.)
  L <- array(1/nc, dim = dim(Y))
  
  list(x = X, y = Y, L = L)
  
}

#' @title Aggregate, an Alternative \link[stats]{formula}-Interface
#' 
#' @description
#' An alternative aggregation function with a \link[stats]{formula}-interface, 
#' to avoid the \link[base]{cbind}-operation in the function \link[stats]{aggregate.formula}.
#' 
#' @param data a \link[base]{data.frame}
#' 
#' @param by a two-sided \link[stats]{formula}
#' 
#' @param ... additional parameters of the function \link[stats]{aggregate.data.frame},
#' *except for* `simplify`
#' 
#' @details
#' The \link[base]{cbind}-operation in the function \link[stats]{aggregate.formula} 
#' messes up with column(s) that are 
#' \describe{
#' \item{\link[base]{factor}}{and treat them as \link[base]{integer}}
#' \item{\link[survival]{Surv}}{and treat them as \link[base]{matrix}}
#' }
#' 
#' The function \link[stats]{aggregate.data.frame} only accepts 
#' a \link[base]{list} of \link[base]{factor}s for the parameter `by`.
#' 
#' Therefore, the function [aggregate2()] is created to take care of 
#' the \link[base]{factor} and \link[survival]{Surv} columns of the input,
#' with a \link[stats]{formula}-interface.
#' 
#' @returns 
#' The function [aggregate2()] returns a \link[base]{data.frame}.
#' 
#' @note
#' The function \link[stats]{aggregate.data.frame} is the workhorse of 
#' the function \link[stats]{aggregate.formula}.
#' 
#' The function \link[spatstat.geom]{as.hyperframe.data.frame}
#' is **designed** to handle the \link[base]{list}-columns 
#' returned by the function \link[stats]{aggregate}.
#' 
#' @importFrom stats aggregate.data.frame model.frame.default update.formula
#' @export
aggregate2 <- function(data, by, ...) {
  
  # drop unused factor levels in all columns of `data`
  data[] <- data |>
    lapply(FUN = \(i) {
      if (!is.factor(i)) return(i)
      factor(i) # drop empty levels!!
    })
  
  if (!is.call(by) || (by[[1L]] != '~') || (length(by) != 3L)) stop('`by` must be two-sided formula')
  
  if (is.symbol(by[[2L]]) && (by[[2L]] == '.')) {
    vars <- names(data)
  } else if (is.call(by[[2L]]) && (by[[2L]][[1L]] == '-')) {
    # e.g. `by = . - x1 - x2 ~ subj_id/image_id`
    vars <- names(data) |>
      lapply(FUN = as.symbol) |>
      Reduce(f = \(e1, e2) call(name = '+', e1, e2)) |>
      call(name = '~', . = _) |>
      eval() |>
      update.formula(new = call(name = '~', by[[2L]])) |>
      all.vars()
  } else {
    vars <- all.vars(by)
    if (!all(vars %in% names(data))) stop()
  }
  
  out <- data[unique(c(all.vars(by[[3L]]), vars))]
  
  f <- by[[3L]] |>
    call(name = '~', . = _) |>
    model.frame.default(formula = _, data = data) |>
    as.list.data.frame() |>
    interaction(drop = TRUE, lex.order = TRUE)
  if (all(table(f) == 1L)) return(out) # exception handling, no aggregation needed!!
  
  z <- out |>
    aggregate.data.frame(
      x = _, 
      by = list(.f = f), ...,
      simplify = TRUE # must!! for `Surv`-column!!
    )
  z[] <- z |> 
    lapply(FUN = unsimplify)
  z <- z[-1L] # grouping structure on the 1st column removed
  return(z)
  
}

#' @title Aggregate Hyper Data Frame
#' 
#' @description
#' To \link[stats]{aggregate} a \link[spatstat.geom]{hyperframe}.
#' 
#' @param x \link[spatstat.geom]{hyperframe}
#' 
#' @param by two-sided \link[stats]{formula}, 
#' whose right-hand-side contains only the regular-column names of the input `x`
#' 
#' @param ... additional parameters of the function \link[stats]{aggregate.data.frame},
#' *except for* `simplify`
#' 
#' @returns 
#' The `S3` method [aggregate.hyperframe()] returns a \link[spatstat.geom]{hyperframe}.
#'  
#' @importFrom spatstat.geom as.hyperframe.data.frame cbind.hyperframe names.hyperframe subset.hyperframe
#' @importFrom stats aggregate aggregate.data.frame
#' @export
aggregate.hyperframe <- function(
    x, 
    by,
    ...
) {
  
  if (!is.call(by) || by[[1L]] != '~' || length(by) != 3L) stop('`by` must be two-sided formula')
  if (!all(all.vars(by[[3L]]) %in% names(unclass(x)$df))) stop('Variables in right-hand-side of `by` must be columns, not hypercolumns')
  
  f <- by[[3L]] |>
    call(name = '~', . = _) |>
    model.frame.default(formula = _, data = unclass(x)$df) |>
    as.list.data.frame() |>
    interaction(drop = TRUE, lex.order = TRUE)
  if (all(table(f) == 1L)) return(x) # exception handling (no need to aggregate)
  
  if (is.symbol(by[[2L]]) && (by[[2L]] == '.')) {
    vars <- names.hyperframe(x)
  } else if (is.call(by[[2L]]) && (by[[2L]][[1L]] == '-')) {
    # e.g. `by = . - x1 - x2 ~ subj_id/image_id`
    vars <- names.hyperframe(x) |>
      lapply(FUN = as.symbol) |>
      Reduce(f = \(e1, e2) call(name = '+', e1, e2)) |>
      call(name = '~', . = _) |>
      eval() |>
      update.formula(new = call(name = '~', by[[2L]])) |>
      all.vars()
  } else {
    vars <- all.vars(by)
    if (!all(vars %in% names.hyperframe(x))) stop()
  }
  
  x. <- x |> 
    subset.hyperframe(
      select = c(all.vars(by[[3L]]), vars) |> 
        unique.default()
    )
  # ?spatstat.geom::`[.hyperframe` # may have bugs..
  # ?spatstat.geom::subset.hyperframe # seems more reliable

  odf <- unclass(x.)$df |>
    aggregate.data.frame(
      x = _, 
      by = list(.f = f), ..., 
      simplify = TRUE # must!! for `Surv`-column!!
    )
  odf[] <- odf |> 
    lapply(FUN = unsimplify)
  odf <- odf[-1L] # grouping structure on the 1st column removed
  
  ohc <- unclass(x.)$hypercolumns |>
    lapply(FUN = split.default, f = f)
  
  ret <- odf |>
    as.hyperframe.data.frame() |>
    list() |>
    c(ohc) |>
    do.call(
      what = cbind.hyperframe, 
      args = _
    )
  
  return(ret)
  
}

#' @title Ki67 Data
#' 
#' @description 
#' Ki67 cell data containing 622 patients.
#' 
#' @format
#' 
#' \describe{
#'   \item{`patientID`}{\link[base]{factor}, unique patient identifier}
#'   \item{`tissueID`}{\link[base]{factor}, TMA core identifier}
#'   \item{`PFS`}{\link[survival]{Surv}, recurrence}
#'   \item{`logKi67`}{\link[base]{double}, log-transformed cell signal intensity of the protein immunofloerscence signal}
#   \item{`x`}{\link[base]{numeric}, \eqn{x}-coordinate in the cell centroid in the TMA core}
#   \item{`y`}{\link[base]{numeric}, \eqn{y}-coordinate in the cell centroid in the TMA core}
#'   \item{`age`}{\link[base]{integer}, age at diagnosis}
#'   \item{`Tstage`}{\link[base]{integer}, tumor stage}
#'   \item{`node`}{\link[base]{factor}, node stage}
#'   \item{`HR`}{\link[base]{logical}, hormone positive status}
#'   \item{`histology`}{\link[base]{integer}, histology grade}
#'   \item{`Her2`}{\link[base]{logical}, Her2 status}
#'   \item{`race`}{\link[base]{character}, race, White, Black, Asian, Native Hawaiian or Other Pacific Islander, American Indian or Alaska Native, Unknown}
#'   \item{`adj_chemo`}{\link[base]{logical}, whether completed adjuvant chemo treatment}
#'   \item{`adj_rad`}{\link[base]{integer}, adjuvant radiation treatment, 0 = unknown,  1 = done, 2 = NOT done}
#' }
#' 
#' @keywords internal
#' @name Ki67
'Ki67'






#' @title wrobel_lung
#' 
#' @description
#' From Dr. Julia Wrobel's data \url{https://sph.emory.edu/profile/faculty/julia-wrobel}.
# \url{http://juliawrobel.com/MI_tutorial/Data/lung.RDA}. # website hacked
#' 
#' @format 
#' A \link[base]{data.frame}
#' 
#' @keywords internal
'wrobel_lung'

# ?cli::cli_text()
# The text to show .. will be concatenated into a single string. 
# Newlines are `not` preserved.

if (FALSE) {
  cli_RPubs_(pub = 'groupedHyperframe')
  cli_QuartoPub_(pub = 'groupedhyperframe')
  cli_QuartoPub_(pub = 'groupedhyperframe', chapter = 'bioinformatics_btaf430')
  cli_Netlify_(pub = 'groupedhyperframe')
  cli_Netlify_(pub = 'groupedhyperframe', chapter = 'bioinformatics_btaf430')
  cli_doi_('10.1002/bimj.4710230408')
}

cli_doi_ <- \(doi) {
  # `x`: 'character' scalar of doi
  sprintf(fmt = '{.href [doi:%s](https://doi.org/%s)}', doi, doi) |>
    cli_text()
}


cli_book_ <- \(
    fmt,
    author = 'tingtingzhan', 
    pub, 
    chapter, print_chapter = TRUE
) {    
  if (missing(chapter) || !print_chapter) {
    paste0('{.url ', fmt, '}') |>
      sprintf(fmt = _, author, pub) |>
      cli_text()
  } else {
    paste0('{.url ', fmt, '/%s.html}') |>
      sprintf(fmt = _, author, pub, chapter) |>
      cli_text()
  }
}



cli_QuartoPub_ <- \(...) cli_book_(fmt = 'https://%s.quarto.pub/%s', ...)

cli_Netlify_ <- \(...) cli_book_(fmt = 'https://%s-%s.netlify.app', ...)

cli_RPubs_ <- \(...) cli_book_(fmt = '{.url https://rpubs.com/%s/%s}', ..., print_chapter = FALSE) 



#' @title Defunct Functions
#' 
#' @description
#' The functions mentioned in hard-copy journals, but later \link[base]{.Defunct}.
#' 
#' @param ... Defunct parameters
#' 
#' @returns 
#' The \link[base]{.Defunct} functions do not have a returned value.
#' 
#' @name defunct
#' @export
aggregate_quantile <- function(...) {
  
  match.call()[[1L]] |> 
    deparse1() |> 
    sprintf(fmt = '%s()') |>
    col_cyan() |> style_bold() |>
    sprintf(fmt = 'Function %s in') |> message()
  cli_doi_('10.1093/bioinformatics/btaf430')
  'has been defunct.\n' |> message()
  
  'Read vignette (mirrors) for details' |> message()
  cli_QuartoPub_(pub = 'groupedhyperframe', chapter = 'bioinformatics_btaf430')
  cli_Netlify_(pub = 'groupedhyperframe', chapter = 'bioinformatics_btaf430')

  .Defunct(new = 'vignette')
  
}

#' @title Vector-List
#' 
#' @description
#' To determine if an R object is a \link[base]{list} of 
#' \link[base]{vector}s with the same 
#' \link[base]{storage.mode},
#' \link[base]{length} and
#' \link[base]{attributes}.
#' 
#' 
#' @param x a \link[stats]{listof}
#' 
#' @param mode \link[base]{character} scalar other than `'any'`, `'complex'` and '`raw`',
#' see the function \link[base]{is.vector}
#' 
#' @returns
#' The function [is.vectorlist()] returns a \link[base]{logical} scalar.
#' 
#' @examples
#' spatstat.data::Kovesi$values |>
#'  is.vectorlist(mode = 'character') |>
#'  stopifnot()
#' spatstat.data::Kovesi$values |>
#'  is.vectorlist(mode = 'numeric')
#' @export
is.vectorlist <- function(
    x, 
    mode = c('logical', 'integer', 'numeric', 'double', 'character')
) {
  
  if (missing(mode)) {
    mode <- x[[1L]] |> 
      storage.mode()
    # mode(1.2) # 'numeric'
    # storage.mode(1.2) # 'double'
  }
  mode <- match.arg(mode)
  
  if (!is.list(x)) return(FALSE)

  id <- x |>
    vapply(FUN = is.vector, mode = mode, FUN.VALUE = NA)
  if (any(!id)) return(FALSE)
  
  id <- x |> 
    lengths(use.names = FALSE) |>
    duplicated.default()
  if (!all(id[-1L])) return(FALSE)
    
  id <- x |>
    #lapply(FUN = names) |> # NULL-name compatible
    lapply(FUN = attributes) |>
    duplicated.default()
  if (!all(id[-1L])) return(FALSE)
  
  return(TRUE)
  
}


#' @title Convert R Object to `vectorlist`
#' 
#' @param x a \link[base]{list}
#' 
#' @param ... additional parameters of the function [is.vectorlist()]
#' 
#' @returns
#' The function [as.vectorlist()] returns an R object of `S3` class `'vectorlist'`.
#' 
#' @examples
#' list(rnorm(6L), rnorm(6L)) |>
#'  as.vectorlist()
#' 
#' @export
as.vectorlist <- function(x, ...) {
  if (!is.vectorlist(x, ...)) stop('input does not qualify as a `vectorlist`')
  class(x) <- c('vectorlist', 'listof', 'list')
  return(x)
}


if (FALSE) {
  library(spatstat)
  methods(class = 'anylist') |>
    attr(which = 'info')
  # tzh don't want/need to use ?spatstat.geom::print.anylist
  # 'vectorlist' does **not** base::inherits from 'anylist'
}




#' @export
print.vectorlist <- function(x, ...) {
  
  '\'vectorlist\'' |>
    col_br_red() |> style_bold() |>
    cat('\n')
  
  x |>
    length() |>
    col_green() |> style_bold() |>
    sprintf(fmt = 'Number of Vectors: %s') |>
    cat('\n')
  
  nm <- x |>
    names()
  if (!all(nm == seq_along(x))) {
    nm |>
      col_cyan() |> style_bold() |>
      paste(collapse = ', ') |>
      sprintf(fmt = 'Name(s): %s') |>
      cat('\n')
  }
  
  x[[1L]] |> 
    storage.mode() |>
    col_blue() |> style_bold() |>
    sprintf(fmt = 'Storage Mode: %s') |>
    cat('\n')
  
  x[[1L]] |> 
    length() |>
    col_magenta() |> style_bold() |>
    sprintf(fmt = 'Individual Vector Length: %s') |>
    cat('\n')
  
}



# @note
# The motivation of 
# the derived class `'vectorlist'` and 
# the method dispatch [t.vectorlist()] 
# is that 
# the `S3` method \link[spatstat.geom]{with.hyperframe}
# could be slow in a batch process.
#' 
# @returns
# The `S3` method dispatch [t.vectorlist()] returns
# a `'vectorlist'` of equi-\link[base]{length}.

#' @importFrom stats setNames
#' @export
t.vectorlist <- function(x) {
  
  # upstream definition to ensure this
  # if (!all(vapply(x, FUN = is.vector, FUN.VALUE = NA))) stop('each element of `x` must be `vector`')
  
  nx <- lengths(x, use.names = FALSE)
  if (!all(duplicated(nx)[-1L])) stop('each element of `x` must be of same `length`')
  
  nm <- lapply(x, FUN = names)
  if (!all(duplicated(nm)[-1L])) stop('each element of `x` must have the same names, or no name')
  
  x |> 
    do.call(what = rbind, args = _) |>
    asplit(MARGIN = 2L, drop = TRUE) |>
    setNames(nm = names(x[[1L]])) |>
    as.vectorlist()

}

#' @title Hyper Data Frame with One-and-Only-One \link[spatstat.geom]{ppp}-Hyper Column
#' 
#' @description
#' To create a hyper data frame with one-and-only-one \link[spatstat.geom]{ppp}-hyper column.
#' 
#' @param marks one-sided \link[stats]{formula}, e.g., 
#' `~ m1+m2`,
#' where \eqn{m_i}'s are one or more \link[spatstat.geom]{marks}
#' 
#' @param coords one-sided \link[stats]{formula}, variable names
#' of the \eqn{x}- and \eqn{y}-coordinates in `data`.
#' Default value is `~x+y`.  
#' 
#' @param by two-sided \link[stats]{formula}
#' 
#' @param data \link[base]{data.frame}
#' 
#' @param window observation window \link[spatstat.geom]{owin}, 
#' default value is the \eqn{x}- and \eqn{y}-span of `coords` in `data`.
#' 
#' @param ... additional parameters of the function \link[spatstat.geom]{ppp}
#' 
#' @returns
#' The function [pppBy()] returns a hyper data frame
#' with one-and-only-one
#' \link[spatstat.geom]{ppp}-hyper column.
#' 
#' @importFrom spatstat.geom owin ppp as.hyperframe.data.frame split.ppp
#' @importFrom stats model.frame.default aggregate.data.frame
#' @export
pppBy <- function(
    marks,
    coords = ~ x + y, 
    by,
    data, 
    window = owin(xrange = range(.x), yrange = range(.y)),
    ...
) {	
  
  # drop unused factor levels in all columns of `x`
  data[] <- data |>
    lapply(FUN = \(i) {
      if (!is.factor(i)) return(i)
      factor(i) # drop empty levels!!
    })
  
  if (is.symbol(by[[2L]]) && (by[[2L]] == '.')) {
    vars <- names(data) |>
      setdiff(y = c(all.vars(marks), all.vars(coords)))
  } else if (is.call(by[[2L]]) && (by[[2L]][[1L]] == '-')) {
    # e.g. `by = . - x1 - x2 ~ subj_id/image_id`
    vars <- names(data) |>
      setdiff(y = c(all.vars(marks), all.vars(coords))) |>
      lapply(FUN = as.symbol) |>
      Reduce(f = \(e1, e2) call(name = '+', e1, e2)) |>
      call(name = '~', . = _) |>
      eval() |>
      update.formula(new = call(name = '~', by[[2L]])) |>
      all.vars()
  } else {
    vars <- all.vars(by)
    if (!all(vars %in% names(data))) stop()
  }
  
  f <- by[[3L]] |> 
    call(name = '~', . = _) |>
    model.frame.default(formula = _, data = data) |>
    as.list.data.frame() |>
    interaction(drop = TRUE, lex.order = TRUE) # one or more hierarchy
  if (all(table(f) == 1L)) stop('shouldnt happen')
  
  d_ag <- data[unique.default(c(all.vars(by[[3L]]), vars))] |>
    aggregate.data.frame(
      x = _,
      by = list(.f = f), 
      FUN = unique_or_identity, 
      simplify = TRUE # must!! for `Surv`-column!!
    )
  d_ag[] <- d_ag |> 
    lapply(FUN = unsimplify)
  hf <- d_ag[-1L] |> # grouping structure on the 1st column removed
    as.hyperframe.data.frame()
  
  xy_ <- as.list.default(coords[[2L]])
  if ((xy_[[1L]] != '+') || (length(xy_) != 3L)) stop('Specify x and y coordinates names as ~x+y')
  if (!is.symbol(x <- xy_[[2L]])) stop('x-coordinates must be a symbol, for now')
  if (!is.symbol(y <- xy_[[3L]])) stop('y-coordinates must be a symbol, for now')
  if (!length(.x <- data[[x]]) || anyNA(.x)) stop('Do not allow missingness in x-coordinates')
  if (!length(.y <- data[[y]]) || anyNA(.y)) stop('Do not allow missingness in y-coordinates')
  
  force(window)
  hf$ppp. <- data[all.vars(marks)] |> # future: use stats::model.frame
    ppp(x = .x, y = .y, window = window, marks = _, ...) |>
    split.ppp(f = f)

  return(hf)
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_doc.R
\docType{data}
\name{Ki67}
\alias{Ki67}
\title{Ki67 Data}
\format{
\describe{
\item{\code{patientID}}{\link[base]{factor}, unique patient identifier}
\item{\code{tissueID}}{\link[base]{factor}, TMA core identifier}
\item{\code{PFS}}{\link[survival]{Surv}, recurrence}
\item{\code{logKi67}}{\link[base]{double}, log-transformed cell signal intensity of the protein immunofloerscence signal}
\item{\code{age}}{\link[base]{integer}, age at diagnosis}
\item{\code{Tstage}}{\link[base]{integer}, tumor stage}
\item{\code{node}}{\link[base]{factor}, node stage}
\item{\code{HR}}{\link[base]{logical}, hormone positive status}
\item{\code{histology}}{\link[base]{integer}, histology grade}
\item{\code{Her2}}{\link[base]{logical}, Her2 status}
\item{\code{race}}{\link[base]{character}, race, White, Black, Asian, Native Hawaiian or Other Pacific Islander, American Indian or Alaska Native, Unknown}
\item{\code{adj_chemo}}{\link[base]{logical}, whether completed adjuvant chemo treatment}
\item{\code{adj_rad}}{\link[base]{integer}, adjuvant radiation treatment, 0 = unknown,  1 = done, 2 = NOT done}
}
}
\usage{
Ki67
}
\description{
Ki67 cell data containing 622 patients.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Defunct.R
\name{defunct}
\alias{defunct}
\alias{aggregate_quantile}
\title{Defunct Functions}
\usage{
aggregate_quantile(...)
}
\arguments{
\item{...}{Defunct parameters}
}
\value{
The \link[base]{.Defunct} functions do not have a returned value.
}
\description{
The functions mentioned in hard-copy journals, but later \link[base]{.Defunct}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_doc.R
\docType{data}
\name{wrobel_lung}
\alias{wrobel_lung}
\title{wrobel_lung}
\format{
A \link[base]{data.frame}
}
\usage{
wrobel_lung
}
\description{
From Dr. Julia Wrobel's data \url{https://sph.emory.edu/profile/faculty/julia-wrobel}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.hyperframe.R
\name{aggregate.hyperframe}
\alias{aggregate.hyperframe}
\title{Aggregate Hyper Data Frame}
\usage{
\method{aggregate}{hyperframe}(x, by, ...)
}
\arguments{
\item{x}{\link[spatstat.geom]{hyperframe}}

\item{by}{two-sided \link[stats]{formula},
whose right-hand-side contains only the regular-column names of the input \code{x}}

\item{...}{additional parameters of the function \link[stats]{aggregate.data.frame},
\emph{except for} \code{simplify}}
}
\value{
The \code{S3} method \code{\link[=aggregate.hyperframe]{aggregate.hyperframe()}} returns a \link[spatstat.geom]{hyperframe}.
}
\description{
To \link[stats]{aggregate} a \link[spatstat.geom]{hyperframe}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pppBy.R
\name{pppBy}
\alias{pppBy}
\title{Hyper Data Frame with One-and-Only-One \link[spatstat.geom]{ppp}-Hyper Column}
\usage{
pppBy(
  marks,
  coords = ~x + y,
  by,
  data,
  window = owin(xrange = range(.x), yrange = range(.y)),
  ...
)
}
\arguments{
\item{marks}{one-sided \link[stats]{formula}, e.g.,
\code{~ m1+m2},
where \eqn{m_i}'s are one or more \link[spatstat.geom]{marks}}

\item{coords}{one-sided \link[stats]{formula}, variable names
of the \eqn{x}- and \eqn{y}-coordinates in \code{data}.
Default value is \code{~x+y}.}

\item{by}{two-sided \link[stats]{formula}}

\item{data}{\link[base]{data.frame}}

\item{window}{observation window \link[spatstat.geom]{owin},
default value is the \eqn{x}- and \eqn{y}-span of \code{coords} in \code{data}.}

\item{...}{additional parameters of the function \link[spatstat.geom]{ppp}}
}
\value{
The function \code{\link[=pppBy]{pppBy()}} returns a hyper data frame
with one-and-only-one
\link[spatstat.geom]{ppp}-hyper column.
}
\description{
To create a hyper data frame with one-and-only-one \link[spatstat.geom]{ppp}-hyper column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0PACKAGE.R
\docType{package}
\name{groupedHyperframe-package}
\alias{groupedHyperframe}
\alias{groupedHyperframe-package}
\title{groupedHyperframe: Grouped Hyper Data Frame}
\description{
To aggregate a hyper data frame, defined in the package 'spatstat.geom', according to a grouping structure. To facilitate downstream analysis based on a "grouped" hyper data frame.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/tingtingzhan/groupedHyperframe}
}

}
\author{
\strong{Maintainer}: Tingting Zhan \email{tingtingzhan@gmail.com} (\href{https://orcid.org/0000-0001-9971-4844}{ORCID})

Authors:
\itemize{
  \item Tingting Zhan \email{tingtingzhan@gmail.com} (\href{https://orcid.org/0000-0001-9971-4844}{ORCID})
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate2.R
\name{aggregate2}
\alias{aggregate2}
\title{Aggregate, an Alternative \link[stats]{formula}-Interface}
\usage{
aggregate2(data, by, ...)
}
\arguments{
\item{data}{a \link[base]{data.frame}}

\item{by}{a two-sided \link[stats]{formula}}

\item{...}{additional parameters of the function \link[stats]{aggregate.data.frame},
\emph{except for} \code{simplify}}
}
\value{
The function \code{\link[=aggregate2]{aggregate2()}} returns a \link[base]{data.frame}.
}
\description{
An alternative aggregation function with a \link[stats]{formula}-interface,
to avoid the \link[base]{cbind}-operation in the function \link[stats]{aggregate.formula}.
}
\details{
The \link[base]{cbind}-operation in the function \link[stats]{aggregate.formula}
messes up with column(s) that are
\describe{
\item{\link[base]{factor}}{and treat them as \link[base]{integer}}
\item{\link[survival]{Surv}}{and treat them as \link[base]{matrix}}
}

The function \link[stats]{aggregate.data.frame} only accepts
a \link[base]{list} of \link[base]{factor}s for the parameter \code{by}.

Therefore, the function \code{\link[=aggregate2]{aggregate2()}} is created to take care of
the \link[base]{factor} and \link[survival]{Surv} columns of the input,
with a \link[stats]{formula}-interface.
}
\note{
The function \link[stats]{aggregate.data.frame} is the workhorse of
the function \link[stats]{aggregate.formula}.

The function \link[spatstat.geom]{as.hyperframe.data.frame}
is \strong{designed} to handle the \link[base]{list}-columns
returned by the function \link[stats]{aggregate}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vectorlist.R
\name{is.vectorlist}
\alias{is.vectorlist}
\title{Vector-List}
\usage{
is.vectorlist(
  x,
  mode = c("logical", "integer", "numeric", "double", "character")
)
}
\arguments{
\item{x}{a \link[stats]{listof}}

\item{mode}{\link[base]{character} scalar other than \code{'any'}, \code{'complex'} and '\code{raw}',
see the function \link[base]{is.vector}}
}
\value{
The function \code{\link[=is.vectorlist]{is.vectorlist()}} returns a \link[base]{logical} scalar.
}
\description{
To determine if an R object is a \link[base]{list} of
\link[base]{vector}s with the same
\link[base]{storage.mode},
\link[base]{length} and
\link[base]{attributes}.
}
\examples{
spatstat.data::Kovesi$values |>
 is.vectorlist(mode = 'character') |>
 stopifnot()
spatstat.data::Kovesi$values |>
 is.vectorlist(mode = 'numeric')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate_FUN.R
\name{aggregate_FUN}
\alias{aggregate_FUN}
\alias{unique_or_identity}
\title{Handy \code{FUN}ctions for \link[stats]{aggregate}}
\usage{
unique_or_identity(x)
}
\arguments{
\item{x}{an R object}
}
\value{
The function \code{\link[=unique_or_identity]{unique_or_identity()}} returns an R object.
}
\description{
Handy functions to be passed to the parameter \code{FUN} of the function \link[stats]{aggregate}
and/or \code{\link[=aggregate2]{aggregate2()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vectorlist.R
\name{as.vectorlist}
\alias{as.vectorlist}
\title{Convert R Object to \code{vectorlist}}
\usage{
as.vectorlist(x, ...)
}
\arguments{
\item{x}{a \link[base]{list}}

\item{...}{additional parameters of the function \code{\link[=is.vectorlist]{is.vectorlist()}}}
}
\value{
The function \code{\link[=as.vectorlist]{as.vectorlist()}} returns an R object of \code{S3} class \code{'vectorlist'}.
}
\description{
Convert R Object to \code{vectorlist}
}
\examples{
list(rnorm(6L), rnorm(6L)) |>
 as.vectorlist()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aug4gam.R
\name{aug4gam}
\alias{aug4gam}
\alias{aug4gam.data.frame}
\alias{aug4gam.hyperframe}
\alias{aug4gam.vectorlist}
\title{Augment Hypercolumn(s) for \link[mgcv]{gam}}
\usage{
aug4gam(x, ...)

\method{aug4gam}{data.frame}(x, ...)

\method{aug4gam}{hyperframe}(x, ...)

\method{aug4gam}{vectorlist}(x, ...)
}
\arguments{
\item{x}{see \strong{Usage}}

\item{...}{additional parameters, currently of no use}
}
\value{
The \code{S3} generic function \code{\link[=aug4gam]{aug4gam()}} returns a \link[base]{data.frame}.
}
\description{
Augment \emph{all} \link[base]{numeric} \link[base]{vector} hypercolumns in a \link[spatstat.geom]{hyperframe}
for \link[mgcv]{gam}.
}
\author{
Tingting Zhan, Erjia Cui
}
